// LED Christmas Tree using AVR ATmega328P.
// Uses Analog Comparator and Timer1 (16 bit) to flash 16 LEDs
// in a Pseudo Random Bit Stream (PRBS) pattern created by
// Linear Feedback Shift Register (LFSR) logic. About 44 mins.
// required for all 65535 combinations at 25 shifts per sec. rate, (LEDdly=40ms)
// Copyright 2020 by Ross Winters, Rev 1.0, 6/10/20

/////////////// !Important! ///////////////////////////////////////////////////////////
// un-comment one of the following lines for your application.
// Default/Development,  16 MHz crystal, CLKcpu= 16MHz, 40 msec. delay
//byte CPUdiv = 0x0; word TIM1cmp = 62500; unsigned int LEDdly = 64000;  // /1 4sec dropout, 38400 baud 
//byte CPUdiv = 0x1; word TIM1cmp = 62500; unsigned int LEDdly = 32000; // /2 testing, 8sec dropout, 19200 baud 
//byte CPUdiv = 0x2; word TIM1cmp = 58594; unsigned int LEDdly = 16000; // /4 testing, 15sec dropout, 9600 baud
//byte CPUdiv = 0x3; word TIM1cmp = 58594; unsigned int LEDdly = 8000; // /8 testing, 30sec dropout, 4800 baud
byte CPUdiv = 0x4; word TIM1cmp = 29297; unsigned int LEDdly = 4000; // /16 testing, 30sec dropout, 2400 baud

// Default, lower power  16 MHz crystal, CLKcpu= 1MHz, 30 sec. dropout, 4 msec. delay
//byte CPUdiv = 0x4; word TIM1cmp = 29297; unsigned int LEDdly = 4000;

// Xmas Tree CLKcpu = 128 KHz, 30 sec. dropout, 40 msec. delay
// Stand alone ATmega328P, 128 KHz internal osc. Requires bootloader fuse programming.
//byte CPUdiv = 0x0; word TIM1cmp = 3750; unsigned int LEDdly = 320 
////////////////////////////////////////////////////////////////////////////////////////  

//#define DEBUG     // un-comment this for serial monitor debug
#ifdef DEBUG
#define LEDPIN 13  // usually 13, slect 0 zero pin, Arduino RX input for testing
#endif

volatile boolean triggered;
volatile boolean rising;
volatile byte acsr_value;
volatile unsigned long int_cntr;  // interrupt counter
byte cyborg_i, peak_i, single_i, twirl_i;
byte c[27][4] = { {0,11,12,13}, {11,12,13,14}, {12,13,14,15}, {13,14,15,16}, {14,15,16,17}, {15,16,17,18}, {16,17,18,19}, {17,18,19,0},
                  {18,19,0,1}, {19,0,1,2}, {0,1,2,3}, {1,2,3,4}, {2,3,4,5}, {3,4,5,8},
                  {8,5,4,3}, {5,4,3,2}, {4,3,2,1}, {3,2,1,0}, {2,1,0,19}, {1,0,19,18}, {0,19,18,17}, {19,18,17,16}, {18,17,16,15}, 
                  {17,16,15,14}, {16,15,14,13}, {15,14,13,12}, {14,13,12,11} };
byte p[17][4] = { {18,8,11,5}, {8,8,12,4}, {8,8,13,3}, {8,8,14,2}, {8,8,15,1}, {8,8,16,0}, {8,8,17,19}, {8,8,18,18}, {18,18,17,19},
                   {11,5,18,18}, {12,4,18,18}, {13,3,18,18}, {14,2,18,18}, {15,1,18,18}, {16,0,18,18}, {17,19,18,18}, {8,8,18,18} };
byte s[16][2] = { {8,11}, {11,12}, {12,13}, {13,14}, {14,15}, {15,16}, {16,17}, {17,18},
                  {18,19}, {19,0}, {0,1}, {1,2}, {2,3}, {3,4}, {4,5}, {5,8} };
byte t[16][2] =  { {4,11}, {11,14}, {14,17}, {17,18}, {18,19}, {19,2}, {2,5}, {5,8},
                  {8,12}, {12,13}, {13,15}, {15,16}, {16,0}, {0,1}, {1,3}, {3,4}};

char D[16], Q[16];
byte Qmap[16] = {11,12,13,14,15,16,17,18,19,0,1,2,3,4,5,8}; // map Q outs to cricuit board LED layout
byte Rmap[8] = {11,14,17,18,19,2,5,8};  // Red LED map
byte Gmap[8] = {12,13,15,16,0,1,3,4};   // Green LED map


ISR(ANALOG_COMP_vect, ISR_BLOCK) { // Analog Comparator Interuppt Service Routine
#ifdef DEBUG
  Serial.println("ISR ANA-COMP Triggered...");
#endif
  if (triggered) return;           // For testing purposes, ignore the interrupt if we haven't reported the last one yet.
                                   // This also effects some crude debouncing.
  cli(); // clear interrupt
  acsr_value = ACSR;
  rising = (acsr_value & bit(ACO)) > 0;
  triggered = true;
  TCNT1 = 0;                            // reset timer1 counter value to 0
  int_cntr>=4000?int_cntr=0:int_cntr++;  // interrupt counter
  if (int_cntr < 2000) { prbs16(); }
  else if (int_cntr < 2500) { cyborg(); }
  else if (int_cntr < 3000) { peak(); }
  else if (int_cntr < 3500) { single(); }
  else { twirl(); }
  
  sei();                          // enable interrupt
}

ISR(TIMER1_COMPA_vect, ISR_BLOCK){   // Timer1 Iterrupt Service Routine. Anytime timer1 triggers, turn off LEDs.
 cli();
 #ifdef DEBUG
 Serial.println("ISR TIMER1 Timeout...");
 #endif

 ACSR |= (0<<ACI);   // disable Analog Comparator interrupt

 close();
 delayMicroseconds(100*LEDdly);  // yes 100*LEDdly

 ACSR |= (1<<ACI);  // enable Analog Comparator interrupt

 sei();
}

void setup() {

#ifdef DEBUG
  Serial.begin (38400); // 38400 for CPUclk /1 /2, /4, /8, /16 testing
  Serial.println ("SETUP()...");
  pinMode(LEDPIN, OUTPUT);
  digitalWrite(LEDPIN, LOW);  // HIGH = lit 
#endif

// setup Analog Comparator
  ADCSRB = ADCSRB & ! bit(ACME);   // disable Analog Comparator Multiplexer Enable
  ACSR =  bit(ACI)                 // clear Analog Comparator Interrupt Flag
        | bit(ACIE)                // set Analog Comparator Interrupt Enable
        | bit(ACIS1) | bit(ACIS0)  // select rising edge: ACIS1/ACIS0 Analog Comparator Interrupt Mode Select
        ;

// Setup Timer1 Interrupt
  TCCR1A = 0;   // set entire TCCR1A register to 0
  TCCR1B = 0;   // same for TCCR1B
  TCNT1  = 0;   //initialize counter value to 0
  OCR1A = 0;   // reset compare A
  OCR1A = TIM1cmp;// = (16*10^6) / (1*1024) - 1 (must be <65536)
  TCCR1B |= (1 << CS12) | (1 << CS10);  // Set CS12 and CS10 bits for 1024 timer1 prescaler 
  TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt
  
// CPU clock divisor and other setup
  CLKPR = (1<<CLKPCE);  // required for system clock prescalar
  CLKPR = CPUdiv;
  
  DIDR1 = DIDR1 | 0x3;    // Disable Digital Input: AIN0, AIN1 for lower power
  prbs_init();

}

void loop() {
if (triggered) show_what_happened();delayMicroseconds(500);
//single();
}

char xnor(char A, char B) {
  if (A == B) { return '1'; } else { return '0'; }
}

void prbs_init() {
  for (byte i = 0; i<16; i++) {
    Q[i]='0';
    D[i]='0';
    pinMode(Qmap[i], OUTPUT);
    }
}

// prbs16 subroutine use taps [16, 5, 3, 2] for xnor feedback i.e. [Q15, Q4, Q2, Q1]
// for maximal length (2^16) - 1 = 65534 (16 bit) combinations.
void prbs16() {

  for (byte i=0; i<16; i++) {
    Q[i] = D[i];
  }

// feedback
  D[0] = xnor(xnor(Q[15], Q[4]), xnor(Q[2], Q[1]));

  for (byte i=0; i<16-1; i++) {
  D[i+1] = Q[i];
  D[i] == '0'? digitalWrite(Qmap[i], HIGH):digitalWrite(Qmap[i], LOW);
  
#ifdef DEBUG
  Serial.print(D[i]);
#endif  
  }
  D[15] == '0'? digitalWrite(Qmap[15], HIGH):digitalWrite(Qmap[15], LOW);
  delayMicroseconds(LEDdly);

#ifdef DEBUG
  Serial.println();
#endif
}

void cyborg() {
  cyborg_i >= 26?cyborg_i=0:cyborg_i++;
  digitalWrite(c[cyborg_i][0], HIGH); // off
  digitalWrite(c[cyborg_i][1], LOW); // on
  digitalWrite(c[cyborg_i][2], LOW); // on
  digitalWrite(c[cyborg_i][3], LOW); // on
  
  delayMicroseconds(LEDdly);
}

void twirl() {
  twirl_i >= 15?twirl_i=0:twirl_i++;
  digitalWrite(t[twirl_i][0], HIGH); // off
  digitalWrite(t[twirl_i][1], LOW); // on
  
  delayMicroseconds(LEDdly);

}

void peak() {
  peak_i >= 16?peak_i=0:peak_i++;
  digitalWrite(p[peak_i][0], HIGH); // off
  digitalWrite(p[peak_i][1], HIGH); // off
  digitalWrite(p[peak_i][2], LOW); // on
  digitalWrite(p[peak_i][3], LOW); // on
  
  delayMicroseconds(LEDdly);
}

void single() {
  single_i >= 15?single_i=0:single_i++;
  digitalWrite(s[single_i][0], HIGH); // off
  digitalWrite(s[single_i][1], LOW); // on
  delayMicroseconds(LEDdly); 
}

void close() {
  for (byte i=0;i<16;i++) {
  digitalWrite(Qmap[i], HIGH); // LED off
  delayMicroseconds(LEDdly);  
  }  
}

// just give an indication via Serial Monitor and LED that we handled an interrupt...
void show_what_happened() {
  triggered = false;

#ifdef DEBUG
  digitalWrite(LEDPIN, HIGH);
  delay(100);
  digitalWrite(LEDPIN, LOW);

  Serial.print ("Triggered! ACSR=");
  Serial.print (acsr_value, BIN);
  if (rising)
    Serial.println (" RISING EDGE (as expected)");
  else
    Serial.println (" FALLING EDGE .. wtf, didn't ask for this, but every so often one slips thru. Maybe a sync issue reading ACSR.");

  Serial.print ("SMCR="); Serial.println (SMCR, BIN);
  Serial.print ("OCR1A="); Serial.println (OCR1A, BIN);
  Serial.print ("TCCR1A="); Serial.println (TCCR1A, BIN);
  Serial.print ("TCCR1B="); Serial.println (TCCR1B, BIN);
  Serial.print ("TCNT1="); Serial.println (TCNT1, BIN);
  Serial.print ("CLKPS3="); Serial.println (bit(CLKPS3), BIN);  // ?
  Serial.print ("CLKPCE="); Serial.println (bit(CLKPCE), BIN);  // ?
  Serial.print ("ADCSRB="); Serial.println (ADCSRB, BIN);
  Serial.print ("ACSR="); Serial.println (ACSR, BIN);
  Serial.print ("DIDR1="); Serial.println (DIDR1, BIN);
  Serial.print ("OSCCAL="); Serial.println (OSCCAL, BIN);
  Serial.print ("CLKPR="); Serial.println (CLKPR, BIN);
  Serial.print ("MCUSR="); Serial.println (MCUSR, BIN);
  Serial.print ("MCUCR="); Serial.println (MCUCR, BIN);
  Serial.print ("PRR="); Serial.println (PRR, BIN);
  Serial.print ("SMCR="); Serial.println (SMCR, BIN);
  Serial.print ("WDTCSR="); Serial.println (WDTCSR, BIN);
#endif
}
